@@ -13,6 +13,8 @@ import json
 
 import jsonschema
 
+from jsonschema.exceptions import RefResolutionError
+
 import dtschema.dtb
 
 schema_base_url = "http://devicetree.org/"
@@ -212,16 +214,23 @@ def _is_matrix_schema(subschema):
 
     return False
 
+
+known_variable_matrix_props = {
+    'fsl,pins',
+}
+
+
 # If we have a matrix with variable inner and outer dimensions, then drop the dimensions
 # because we have no way to reconstruct them.
-def _fixup_int_matrix(subschema):
+def _fixup_int_matrix(propname, subschema):
     if not _is_matrix_schema(subschema):
         return
 
     outer_dim = _get_array_range(subschema)
     inner_dim = _get_array_range(subschema.get('items', {}))
 
-    if outer_dim[0] != outer_dim[1] and inner_dim[0] != inner_dim[1]:
+    if propname in known_variable_matrix_props or \
+       (outer_dim[0] != outer_dim[1] and inner_dim[0] != inner_dim[1]):
         subschema.pop('items', None)
         subschema.pop('maxItems', None)
         subschema.pop('minItems', None)
@@ -229,10 +238,17 @@ def _fixup_int_matrix(subschema):
 
 
 int_array_re = re.compile('int(8|16|32|64)-array')
-unit_types_re = re.compile('-(bits|percent|mhz|hz|sec|ms|us|ns|ps|mm|microamp|microamp-hours|ohms|micro-ohms|microwatt-hours|microvolt|picofarads|celsius|millicelsius|kpascal)$')
+unit_types_re = re.compile('-(kBps|bits|percent|bp|m?hz|sec|ms|us|ns|ps|mm|nanoamp|(micro-)?ohms|micro(amp|watt)(-hours)?|milliwatt|microvolt|picofarads|(milli)?celsius|kelvin|kpascal)$')
 
 # Remove this once we remove array to matrix fixups
-known_array_props = {'assigned-clock-rates', 'linux,keycodes'}
+known_array_props = {
+    'assigned-clock-rates',
+    'linux,keycodes',
+    'max8997,pmic-buck1-dvs-voltage',
+    'max8997,pmic-buck2-dvs-voltage',
+    'max8997,pmic-buck5-dvs-voltage',
+}
+
 
 def is_int_array_schema(propname, subschema):
     if 'allOf' in subschema:
@@ -416,7 +432,7 @@ def fixup_vals(propname, schema):
 
     _fixup_reg_schema(propname, schema)
     _fixup_remove_empty_items(schema)
-    _fixup_int_matrix(schema)
+    _fixup_int_matrix(propname, schema)
     _fixup_int_array_min_max_to_matrix(propname, schema)
     _fixup_int_array_items_to_matrix(propname, schema)
     _fixup_string_to_array(propname, schema)
@@ -462,6 +478,11 @@ def fixup_sub_schema(schema, is_prop):
     if is_prop:
         fixup_node_props(schema)
 
+    # 'additionalProperties: true' doesn't work with 'unevaluatedProperties', so
+    # remove it. It's in the schemas for common (incomplete) schemas.
+    if 'additionalProperties' in schema and schema['additionalProperties'] == True:
+        schema.pop('additionalProperties', None)
+
     for k, v in schema.items():
         if k in ['select', 'if', 'then', 'else', 'additionalProperties', 'not']:
             fixup_sub_schema(v, False)
@@ -491,6 +512,7 @@ def fixup_node_props(schema):
     schema.setdefault('properties', dict())
     schema['properties'].setdefault('phandle', True)
     schema['properties'].setdefault('status', True)
+    schema['properties'].setdefault('secure-status', True)
     schema['properties'].setdefault('$nodename', True)
 
     keys = list()
@@ -501,10 +523,10 @@ def fixup_node_props(schema):
         keys.extend(schema['patternProperties'])
 
     for key in keys:
-        if "pinctrl" in key:
+        if re.match(r'^pinctrl-[0-9]', key):
             break
     else:
-        schema['properties']['pinctrl-names'] = True
+        schema['properties'].setdefault('pinctrl-names', True)
         schema.setdefault('patternProperties', dict())
         schema['patternProperties']['pinctrl-[0-9]+'] = True
 
@@ -653,7 +675,7 @@ def fixup_interrupts(schema):
 def make_compatible_schema(schemas):
     compat_sch = [{'enum': []}]
     compatible_list = set()
-    for sch in schemas:
+    for sch in schemas.values():
         compatible_list |= extract_compatibles(sch)
 
     # Allow 'foo' values for examples
@@ -663,13 +685,13 @@ def make_compatible_schema(schemas):
     for c in compatible_list:
         if prog.match(c):
             # Exclude the generic pattern
-            if c != '^[a-zA-Z][a-zA-Z0-9,+\-._]+$':
+            if c != '^[a-zA-Z0-9][a-zA-Z0-9,+\-._]+$':
                 compat_sch += [{'pattern': c}]
         else:
             compat_sch[0]['enum'].append(c)
 
     compat_sch[0]['enum'].sort()
-    schemas += [{
+    schemas['generated-compatibles'] = {
         '$id': 'generated-compatibles',
         '$filename': 'Generated schema of documented compatible strings',
         'select': True,
@@ -680,7 +702,19 @@ def make_compatible_schema(schemas):
                 }
             }
         }
-    }]
+    }
+
+
+def get_undocumented_compatibles(compatible_list):
+    global schema_cache
+    undoc_compats = []
+
+    validator = dtschema.DTValidator(schema_cache['generated-compatibles'])
+    for compat in compatible_list:
+        if not validator.is_valid({"compatible": [ compat ]}):
+            undoc_compats += [ compat ]
+
+    return undoc_compats
 
 
 def process_schema(filename):
@@ -698,6 +732,10 @@ def process_schema(filename):
               file=sys.stderr)
         #print(exc.message)
         return
+    except:
+        print(filename + ": ignoring, unknown error in schema (not a schema?)",
+              file=sys.stderr)
+        return
 
     if 'select' not in schema:
         print(filename + ": warning: no 'select' found in schema found", file=sys.stderr)
@@ -709,8 +747,7 @@ def process_schema(filename):
 
 
 def process_schemas(schema_paths, core_schema=True):
-    ids = []
-    schemas = []
+    schemas = {}
 
     for filename in schema_paths:
         if not os.path.isfile(filename):
@@ -718,10 +755,10 @@ def process_schemas(schema_paths, core_schema=True):
         sch = process_schema(os.path.abspath(filename))
         if not sch:
             continue
-        schemas.append(sch)
-        if ids.count(sch['$id']):
+        if sch['$id'] in schemas:
             print(os.path.abspath(filename) + ": duplicate '$id' value '" + sch['$id'] + "'", file=sys.stderr)
-        ids.append(sch['$id'])
+        else:
+            schemas[sch['$id']] = sch
 
     if core_schema:
         schema_paths.append(os.path.join(schema_basedir, 'schemas/'))
@@ -735,10 +772,10 @@ def process_schemas(schema_paths, core_schema=True):
             sch = process_schema(os.path.abspath(filename))
             if sch:
                 count += 1
-                schemas.append(sch)
-                if ids.count(sch['$id']):
+                if sch['$id'] in schemas:
                     print(os.path.abspath(filename) + ": duplicate '$id' value '" + sch['$id'] + "'", file=sys.stderr)
-                ids.append(sch['$id'])
+                else:
+                    schemas[sch['$id']] = sch
 
         if count == 0:
             print("warning: no schema found in path: %s" % path, file=sys.stderr)
@@ -779,10 +816,13 @@ def _merge_dim(dim1, dim2):
 type_re = re.compile('(flag|u?int(8|16|32|64)(-(array|matrix))?|string(-array)?|phandle(-array)?)')
 
 
-def _extract_prop_type(props, schema, propname, subschema):
+def _extract_prop_type(props, schema, propname, subschema, is_pattern):
     if not isinstance(subschema, dict):
         return
 
+    if propname.startswith('$'):
+        return
+
     if subschema.keys() & {'properties', 'patternProperties', 'additionalProperties'}:
         _extract_subschema_types(props, schema, subschema)
 
@@ -793,106 +833,193 @@ def _extract_prop_type(props, schema, propname, subschema):
         for p in sch_path:
             subschema = subschema[p]
         #print(propname, sch_path, subschema, file=sys.stderr)
-        _extract_prop_type(props, schema, propname, subschema)
-
-    try:
-        prop_type = type_re.search(subschema['$ref']).group(0)
-    except:
-        if 'type' in subschema and subschema['type'] == 'boolean':
-            prop_type = 'flag'
-        elif 'items' in subschema:
-            items = subschema['items']
-            if (isinstance(items, list) and _is_string_schema(items[0])) or \
-               (isinstance(items, dict) and _is_string_schema(items)):
-                # implicit string type
-                prop_type = 'string-array'
-            else:
-                prop_type = None
-        else:
-            prop_type = None
-
-    if prop_type:
-        props.setdefault(propname, {})
-
-        if prop_type == 'phandle-array' or prop_type.endswith('-matrix'):
-            dim = (_get_array_range(subschema), _get_array_range(subschema.get('items', {})))
-            if dim != ((0, 0), (0, 0)):
-                if not 'dim' in props[propname]:
-                    props[propname]['dim'] = dim
-                elif props[propname]['dim'] != dim:
-                    # Conflicting dimensions
-                    props[propname]['dim'] = _merge_dim(props[propname]['dim'], dim)
-
-        if 'type' in props[propname]:
-            if prop_type in props[propname]['type']:
-                # Already have the same type
-                return
-            for t in props[propname]['type']:
-                if prop_type in t:
-                    # Already have the looser type
-                    return
-                if t in prop_type:
-                    # Replace scalar type with array type
-                    i = props[propname]['type'].index(t)
-                    props[propname]['type'][i] = prop_type
-                    props[propname]['$id'][i] = schema['$id']
-                    return
-        else:
-            props[propname]['type'] = []
-            props[propname]['$id'] = []
-
-        props[propname]['type'] += [ prop_type ]
-        props[propname]['$id'] += [ schema['$id'] ]
-        return
+        _extract_prop_type(props, schema, propname, subschema, is_pattern)
 
     for k in subschema.keys() & {'allOf', 'oneOf', 'anyOf'}:
         for v in subschema[k]:
-            _extract_prop_type(props, schema, propname, v)
+            _extract_prop_type(props, schema, propname, v, is_pattern)
+
+    props.setdefault(propname, [])
+
+    new_prop = {}
+    prop_type = None
+
+    if 'type' in subschema and subschema['type'] == 'object':
+        prop_type = 'node'
+    else:
+        try:
+            prop_type = type_re.search(subschema['$ref']).group(0)
+        except:
+            if 'type' in subschema and subschema['type'] == 'boolean':
+                prop_type = 'flag'
+            elif 'items' in subschema:
+                items = subschema['items']
+                if (isinstance(items, list) and _is_string_schema(items[0])) or \
+                (isinstance(items, dict) and _is_string_schema(items)):
+                    # implicit string type
+                    prop_type = 'string-array'
+                elif not (isinstance(items, list) and len(items) == 1 and \
+                    'items' in items and isinstance(items['items'], list) and len(items['items']) == 1) and \
+                    unit_types_re.search(propname):
+                    prop_type = 'uint32-matrix'
+                else:
+                    prop_type = None
+            elif '$ref' in subschema and re.search(r'\.yaml#?$', subschema['$ref']):
+                prop_type = 'node'
+            else:
+                prop_type = None
+
+    new_prop['type'] = prop_type
+    new_prop['$id'] = [schema['$id']]
+    if is_pattern:
+        new_prop['regex'] = re.compile(propname)
+
+    if not prop_type:
+        if len(props[propname]) == 0:
+            props[propname] += [new_prop]
+        return
+
+    # handle matrix dimensions
+    if prop_type == 'phandle-array' or prop_type.endswith('-matrix'):
+        dim = (_get_array_range(subschema), _get_array_range(subschema.get('items', {})))
+        new_prop['dim'] = dim
+    else:
+        dim = ((0, 0), (0, 0))
+
+    if propname in props:
+        dup_prop = None
+        for p in props[propname]:
+            if p['type'] is None:
+                dup_prop = p
+                break
+            if dim != ((0, 0), (0, 0)) and (p['type'] == 'phandle-array' or p['type'].endswith('-matrix')):
+                if not 'dim' in p:
+                    p['dim'] = dim
+                elif p['dim'] != dim:
+                    # Conflicting dimensions
+                    p['dim'] = _merge_dim(p['dim'], dim)
+                return
+            if p['type'].startswith(prop_type):
+                # Already have the same or looser type
+                if schema['$id'] not in p['$id']:
+                    p['$id'] += [schema['$id']]
+                return
+            elif p['type'] in prop_type:
+                # Replace scalar type with array type
+                new_prop['$id'] += p['$id']
+                dup_prop = p
+                break
+
+        if dup_prop:
+            props[propname].remove(dup_prop)
+
+    props[propname] += [new_prop]
 
 
 def _extract_subschema_types(props, schema, subschema):
     if not isinstance(subschema, dict):
         return
 
-    for k in subschema.keys() & {'properties', 'patternProperties', 'additionalProperties'}:
+    if 'additionalProperties' in subschema:
+        _extract_subschema_types(props, schema, subschema['additionalProperties'])
+
+    for k in subschema.keys() & {'properties', 'patternProperties'}:
         if isinstance(subschema[k], dict):
             for p,v in subschema[k].items():
-                _extract_prop_type(props, schema, p, v)
+                _extract_prop_type(props, schema, p, v, k == 'patternProperties')
 
 
-def extract_types(schemas):
-    if not isinstance(schemas, list):
-        return
+def extract_types():
+    global schema_cache
 
     props = {}
-    for sch in schemas:
+    for sch in schema_cache.values():
         _extract_subschema_types(props, sch, sch)
 
     return props
 
 
-def get_prop_types():
-    global schema_cache
+def get_prop_types(want_missing_types=False, want_node_types=False):
     pat_props = {}
 
-    props = dtschema.extract_types(schema_cache)
+    props = dtschema.extract_types()
 
-    # hack to remove aliases pattern
+    # hack to remove aliases and generic patterns
     del props['^[a-z][a-z0-9\-]*$']
+    props.pop('^[a-zA-Z][a-zA-Z0-9\\-_]{0,63}$', None)
+    props.pop('^.*$', None)
+    props.pop('.*', None)
+
+    # Remove node types
+    if not want_node_types:
+        for val in props.values():
+            val[:] = [t for t in val if t['type'] != 'node']
 
     # Remove all properties without a type
-    for key in [key for key in props if 'type' not in props[key] ]: del props[key]
+    if not want_missing_types:
+        for val in props.values():
+            val[:] = [t for t in val if t['type'] is not None]
+
+    # Delete any entries now empty due to above operations
+    for key in [key for key in props if len(props[key]) == 0]: del props[key]
 
     # Split out pattern properties
-    for key in [key for key in props if re.fullmatch('(^\^.*|.*\$$|.*[*[].*)', key) ]:
-        #print(key, props[key])
-        pat_props[key] = props[key]
-        pat_props[key]['regex'] = re.compile(key)
+    for key in [key for key in props if len(props[key]) and 'regex' in props[key][0] ]:
+        # Only want patternProperties with type and some amount of fixed string
+        if re.search(r'[0-9a-zA-F-]{3}', key):
+            #print(key, props[key], file=sys.stderr)
+            pat_props[key] = props[key]
         del props[key]
 
     return [ props, pat_props ]
 
 
+props = None
+pat_props = None
+
+
+def property_get_type(propname):
+    global props
+    global pat_props
+
+    if not props:
+        props, pat_props = get_prop_types()
+
+    type = []
+    if propname in props:
+        for v in props[propname]:
+            if v['type']:
+                type += [v['type']]
+    if len(type) == 0:
+        for v in pat_props.values():
+            if v[0]['type'] and v[0]['type'] not in type and v[0]['regex'].search(propname):
+                type += [v[0]['type']]
+
+    # Don't return 'node' as a type if there's other types
+    if len(type) > 1 and 'node' in type:
+        type.remove('node')
+    return type
+
+
+def property_get_type_dim(propname):
+    global props
+    global pat_props
+
+    if not props:
+        props, pat_props = get_prop_types()
+
+    if propname in props:
+        for v in props[propname]:
+            if 'dim' in v:
+                return v['dim']
+
+    for v in pat_props.values():
+        if v[0]['type'] and 'dim' in v[0] and v[0]['regex'].search(propname):
+            return v[0]['dim']
+
+    return None
+
+
 def load(filename, line_number=False):
     if filename.endswith('.dtb'):
         with open(filename, 'rb') as f:
@@ -904,25 +1031,87 @@ def load(filename, line_number=False):
             return yaml.load(f.read())
 
 
-schema_cache = []
-
-
-def set_schema(schemas):
+def make_property_type_cache():
     global schema_cache
-    schema_cache = schemas
 
+    props, pat_props = get_prop_types()
+
+    for val in props.values():
+        val[:] = [t for t in val if 'regex' not in t]
+        for t in val: del t['$id']
+
+    schema_cache['generated-types'] = {
+        '$id': 'generated-types',
+        '$filename': 'Generated property types',
+        'select': False,
+        'properties': props
+    }
+
+    for val in pat_props.values():
+        for t in val:
+            t.pop('regex', None)
+            del t['$id']
+
+    schema_cache['generated-pattern-types'] = {
+        '$id': 'generated-pattern-types',
+        '$filename': 'Generated property types',
+        'select': False,
+        'properties': pat_props
+    }
+
+
+schema_cache = {}
+
+def set_schemas(schema_files, core_schema=True):
+    global schema_cache, pat_props, props
+
+    if len(schema_files) == 1 and os.path.isfile(schema_files[0]):
+        # a processed schema file
+        schema_cache = dtschema.load_schema(os.path.abspath(schema_files[0]))
+        # Convert old format to new
+        if isinstance(schema_cache, list):
+            d = {}
+            for sch in schema_cache:
+                d[sch['$id']] = sch
+            schema_cache = d
+
+        if 'generated-types' in schema_cache:
+            props = schema_cache['generated-types']['properties']
+        if 'generated-pattern-types' in schema_cache:
+            pat_props = schema_cache['generated-pattern-types']['properties']
+            for k in pat_props:
+                pat_props[k][0]['regex'] = re.compile(k)
+    else:
+        schema_cache = process_schemas(schema_files, core_schema)
+        make_property_type_cache()
+
+    return schema_cache
 
 def http_handler(uri):
     global schema_cache
     '''Custom handler for http://devicetre.org YAML references'''
     try:
         if schema_base_url in uri:
-            for sch in schema_cache:
-                if uri in sch['$id']:
-                    return sch
+            my_uri = uri + '#'
+            if my_uri in schema_cache:
+                return schema_cache[my_uri]
+            # If we have a schema_cache, then the schema should have been there unless the schema had errors
+            if len(schema_cache):
+                return False
             if 'meta-schemas' in uri:
                 return load_schema(uri.replace(schema_base_url, ''))
-            return process_schema(uri.replace(schema_base_url, ''))
+
+            try:
+                schema = load_schema(uri.replace(schema_base_url, ''))
+            except Exception as exc:
+                raise RefResolutionError('Unable to find schema file matching $id: ' + uri)
+
+            try:
+                DTValidator.check_schema(schema)
+            except Exception as exc:
+                raise RefResolutionError('Error in referenced schema matching $id: ' + uri)
+
+            return schema
 
         from urllib.request import urlopen
 
